#include "detect.h"

#ifdef __linux__
#include <poll.h>
#include <libudev.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif/*__linux__*/

#ifdef _WIN32
#define WINVER 0x502

#include <windows.h>

#include <initguid.h>
#include <ddk/usbiodef.h>
#include <Setupapi.h>

#include <tchar.h>
#include <conio.h>
#include <dbt.h>
#include <stdio.h>

#endif/*_WIN32*/

#ifdef __linux__
struct dev_list_t {
	char *path;
	unsigned short vid;
	unsigned short pid;
	struct dev_list_t *next;
};
#endif/*__linux__*/

struct detect_t {
#ifdef __linux__
	struct udev* hotplug;
	struct udev_monitor* hotplug_monitor;
	struct dev_list_t *devlist;
#endif/*__linux__*/

#ifdef _WIN32
	HWND hwnd;
	HDEVNOTIFY hDeviceNotify;
#endif/*_WIN32*/

	detect_hotplug_cb_fn attach;
  detect_hotplug_cb_fn detach;
	void *user_data;
};

#ifdef __linux__
static void dev_list_add(struct detect_t *h, const char *path,
												 unsigned short vid, unsigned short pid)
{
	struct dev_list_t *dev =
		(struct dev_list_t*)malloc(sizeof(struct dev_list_t));
	dev->path = strdup(path);
	dev->vid = vid;
	dev->pid = pid;
	dev->next = NULL;

	struct dev_list_t *p = h->devlist;
	if(!p) {
		h->devlist = dev;
		return;
	}

	while(p->next) {
		p = p->next;
	}
	
	p->next = dev;
}

static void dev_list_remove(struct detect_t *h, const char *path)
{
	struct dev_list_t *p = h->devlist;
	if(!p) return;

	if(!strcmp(p->path, path)) {
		h->devlist = p->next;
		free(p->path);
		free(p);
		return;
	}

	while(p->next) {
		if(!strcmp(p->next->path, path)) {
			struct dev_list_t *pp = p->next;
			p->next = pp->next;
			free(pp->path);
			free(pp->next);
			free(pp);
			return;
		}
		p = p->next;
	}
	
	printf("Not found (remove)...!\n");
}

static void dev_list_find(struct detect_t *h, const char *path,
													unsigned short *vid, unsigned short *pid)
{
	struct dev_list_t *p = h->devlist;
	while(p) {
		if(!strcmp(p->path, path)) {
			*vid = p->vid;
			*pid = p->pid;
			return;
		}
		p = p->next;
	}
	
	printf("Not found (find)...!\n");
}
#endif/*__linux__*/


#ifdef _WIN32
LRESULT CALLBACK WinProcCallback(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	struct detect_t *h = (struct detect_t*)GetWindowLong(hwnd, GWL_USERDATA);
	(void)h;

	switch(msg) {
	case WM_DEVICECHANGE:
		{
			struct detect_device_t *device = NULL;

			PDEV_BROADCAST_HDR phdr = (PDEV_BROADCAST_HDR)lp;

			if(phdr && phdr->dbch_devicetype == DBT_DEVTYP_DEVICEINTERFACE) {
				PDEV_BROADCAST_DEVICEINTERFACE lpDevIf =
					(PDEV_BROADCAST_DEVICEINTERFACE)lp;

        char* dbcc_name = strdup(lpDevIf->dbcc_name);
        HDEVINFO devInfoList = SetupDiCreateDeviceInfoList( NULL, NULL );
        
        SP_DEVICE_INTERFACE_DATA devInterfaceData;
        ZeroMemory( &devInterfaceData, sizeof( devInterfaceData ) );
        devInterfaceData.cbSize = sizeof( devInterfaceData );
        BOOL b = SetupDiOpenDeviceInterface( devInfoList,
                                             dbcc_name,
                                             0,
                                             &devInterfaceData );
				(void)b;
        free(dbcc_name);
        
        DWORD required;
        SP_DEVICE_INTERFACE_DETAIL_DATA devDetailData;
        ZeroMemory( &devDetailData, sizeof( devDetailData ) );
        devDetailData.cbSize = sizeof( devDetailData );

				SP_DEVINFO_DATA devInfoData;
        ZeroMemory( &devInfoData, sizeof( devInfoData ) );
        devInfoData.cbSize = sizeof( devInfoData );
        b = SetupDiGetDeviceInterfaceDetail( devInfoList,
                                         &devInterfaceData,
                                         &devDetailData,
                                         sizeof( devDetailData ),
                                         &required,
                                         &devInfoData );
       
        TCHAR deviceId[1000];
        b = SetupDiGetDeviceInstanceId( devInfoList,
                                        &devInfoData,
                                        deviceId,
                                        sizeof(deviceId),
                                        NULL );

				char *vid = strstr(deviceId, "VID_");
				if(vid != NULL) vid += 4;

				char *pid = strstr(deviceId, "PID_");
				if(pid != NULL)	pid += 4;

				if(pid) {
					pid[4] = '\0';
					if(!device) {
						device =
							(struct detect_device_t*)malloc(sizeof(struct detect_device_t));
					}
					device->idProduct = strtol(pid, NULL, 16); 
				}

				if(vid) {
					vid[4] = '\0';
					if(!device) {
						device =
							(struct detect_device_t*)malloc(sizeof(struct detect_device_t));
					}
					device->idVendor = strtol(vid, NULL, 16);
				}
			}

			switch(wp) {
			case DBT_DEVICEARRIVAL:
				if(h->attach) h->attach(device, h->user_data);
				break;
			case DBT_DEVICEREMOVECOMPLETE:
				if(h->detach) h->detach(device, h->user_data);
				break;
			case DBT_DEVNODES_CHANGED:
			default:
				break;
			}

			if(device) free(device);
		}
		break;
	default:
		break;
	}
	
	return DefWindowProc(hwnd, msg, wp, lp);
}
#endif/*OS_WINDOWS*/

int detect_init(struct detect_t **handle)
{
	struct detect_t *h = (struct detect_t *)malloc(sizeof(struct detect_t));

	h->attach = NULL;
	h->detach = NULL;
	h->user_data = NULL;

#ifdef __linux__
	h->devlist = NULL;

  // create the udev object
  h->hotplug = udev_new();
  if(!h->hotplug)
  {
    printf("Cannot create udev object\n");
		free(h);
		return 1;
  }

  // create the udev monitor
  h->hotplug_monitor = udev_monitor_new_from_netlink(h->hotplug, "udev");

  // start receiving hotplug events
	udev_monitor_filter_add_match_subsystem_devtype(h->hotplug_monitor,
																									"usb", "usb_device");
  udev_monitor_enable_receiving(h->hotplug_monitor);

	struct udev_enumerate *de = udev_enumerate_new (h->hotplug);
	udev_enumerate_add_match_subsystem(de, "usb");
	udev_enumerate_scan_devices(de);

	struct udev_list_entry *lst = udev_enumerate_get_list_entry(de);
	while(lst) {
		struct udev_device *dev =
			udev_device_new_from_syspath(h->hotplug,
																	 udev_list_entry_get_name(lst));

		if(udev_device_get_devnode(dev)) {
			unsigned short idVendor =
				strtol(udev_device_get_sysattr_value(dev, "idVendor"), NULL, 16); 
			unsigned short idProduct =
				strtol(udev_device_get_sysattr_value(dev, "idProduct"), NULL, 16); 

			dev_list_add(h, udev_device_get_devnode(dev), idVendor, idProduct);
		}

		udev_device_unref(dev);

		lst = udev_list_entry_get_next(lst);
	}

	udev_enumerate_unref(de);

#endif/*__linux__*/

#ifdef _WIN32
	WNDCLASSEX wcex;

	memset(&wcex, 0, sizeof(wcex));
		wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.lpfnWndProc = WinProcCallback;
	wcex.hInstance = GetModuleHandle(NULL);	
	wcex.lpszClassName = "UsbHotplugClass";
	wcex.cbWndExtra = sizeof(struct detect_t*); // Size of data.

	RegisterClassEx(&wcex);

	h->hwnd =
		CreateWindowEx(0, wcex.lpszClassName, "UsbHotplug",	 0, 0, 0, 0,
									 0,	 0, NULL, GetModuleHandle(NULL), NULL);

	SetWindowLong(h->hwnd, GWL_USERDATA, (LONG)h);


	DEV_BROADCAST_DEVICEINTERFACE *filter =
	(DEV_BROADCAST_DEVICEINTERFACE*)malloc(sizeof(DEV_BROADCAST_DEVICEINTERFACE));

	memset(filter, 0, sizeof(DEV_BROADCAST_DEVICEINTERFACE));
	filter->dbcc_size = sizeof(DEV_BROADCAST_DEVICEINTERFACE);
	filter->dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
	filter->dbcc_classguid = GUID_DEVINTERFACE_USB_DEVICE;

	h->hDeviceNotify =
		RegisterDeviceNotification(h->hwnd, filter, DEVICE_NOTIFY_WINDOW_HANDLE);
	
	if(h->hDeviceNotify == 0) {
		//printf("RegisterDeviceNotification error\n");
		free(h);
		return 1;
	}
#endif/*_WIN32*/

	*handle = h;
	return 0;
}

void detect_exit(struct detect_t *h)
{
#ifdef __linux__
  // destroy the udev monitor
  udev_monitor_unref(h->hotplug_monitor);

  // destroy the udev object
  udev_unref(h->hotplug);
#endif/*__linux__*/

#ifdef _WIN32
	UnregisterDeviceNotification(h->hDeviceNotify);
	// TODO: Clean up hDeviceNotify, window class and window handle.
#endif/*_WIN32*/

	free(h);
}

int detect_handle_events_timeout(struct detect_t *h, struct timeval *tv)
{
	int ms = tv->tv_sec * 1000 + tv->tv_usec / 1000;

#ifdef __linux__
  // create the poll item
  struct pollfd items[1];
  items[0].fd = udev_monitor_get_fd(h->hotplug_monitor);
  items[0].events = POLLIN;
  items[0].revents = 0;

  // while there are hotplug events to process
  while(poll(items, 1, ms) > 0) {		
    // receive the relevant device
    struct udev_device* dev = udev_monitor_receive_device(h->hotplug_monitor);
    if(!dev) {
      // error receiving device, skip it
      continue;
    }

		if(!strcmp(udev_device_get_action(dev), "add") && h->attach) {
			struct detect_device_t *device =
				(struct detect_device_t *)malloc(sizeof(struct detect_device_t));
			
			device->idVendor =
				strtol(udev_device_get_sysattr_value(dev, "idVendor"), NULL, 16); 
			device->idProduct =
				strtol(udev_device_get_sysattr_value(dev, "idProduct"), NULL, 16); 

			dev_list_add(h, udev_device_get_devnode(dev),
									 device->idVendor, device->idProduct);

			h->attach(device, h->user_data);
			free(device);
		}
		if(!strcmp(udev_device_get_action(dev), "remove") && h->detach) {
			struct detect_device_t *device =
				(struct detect_device_t *)malloc(sizeof(struct detect_device_t));

			dev_list_find(h, udev_device_get_devnode(dev),
										&device->idVendor, &device->idProduct);

			dev_list_remove(h, udev_device_get_devnode(dev));

			h->detach(device, h->user_data);
			free(device);
		}

		// destroy the relevant device
		udev_device_unref(dev);

    // clear the revents
    items[0].revents = 0;
  }
#endif/*__linux__*/

#ifdef _WIN32
	UINT_PTR timer = SetTimer(h->hwnd, 0, ms, NULL);

	MSG msg; 
	int ret = GetMessage(&msg, NULL, 0, 0);

	if(ret <= 0) return 0;
	
	TranslateMessage(&msg);
	DispatchMessage(&msg);

	KillTimer(h->hwnd, timer);
#endif/*_WIN32*/

	return 0;
}

void detect_register_hotplug_listeners(struct detect_t *handle,
																			 detect_hotplug_cb_fn connected_cb,
																			 detect_hotplug_cb_fn disconnected_cb,
																			 void *user_data)
{
	handle->attach = connected_cb;
	handle->detach = disconnected_cb;
	handle->user_data = user_data;
}
